// Resource-to-File 1.1 - CLI edition
// by Gabriele Biffi - www.biffuz.it
// Distributed under the terms of the MIT license.

// Standard C headers
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// BeOS headers
#include <ByteOrder.h>
#include <File.h>
#include <Resources.h>
#include <TypeConstants.h>

// Type Codes ==================================================================

typedef struct {
	const char *name;
	type_code code;
} TypeCode;

#define TYPE_CODES_COUNT 40
const TypeCode kTypeCodes[] = {
	{"B_ANY_TYPE",						'ANYT'}, 
	{"B_BOOL_TYPE",						'BOOL'}, 
	{"B_CHAR_TYPE",						'CHAR'}, 
	{"B_COLOR_8_BIT_TYPE",				'CLRB'}, 
	{"B_DOUBLE_TYPE",					'DBLE'}, 
	{"B_FLOAT_TYPE",					'FLOT'}, 
	{"B_FONT_TYPE",						'FONt'}, 
	{"B_GRAYSCALE_8_BIT_TYPE",			'GRYB'}, 
	{"B_INT64_TYPE",					'LLNG'}, 
	{"B_INT32_TYPE",					'LONG'}, 
	{"B_INT16_TYPE",					'SHRT'}, 
	{"B_INT8_TYPE",						'BYTE'}, 
	{"B_MESSAGE_TYPE",					'MSGG'}, 
	{"B_MESSENGER_TYPE",				'MSNG'}, 
	{"B_MIME_TYPE",						'MIME'}, 
	{"B_MONOCHROME_1_BIT_TYPE",			'MNOB'}, 
	{"B_OBJECT_TYPE",					'OPTR'}, 
	{"B_OFF_T_TYPE",					'OFFT'}, 
	{"B_PATTERN_TYPE",					'PATN'}, 
	{"B_POINTER_TYPE",					'PNTR'}, 
	{"B_POINT_TYPE",					'BPNT'}, 
	{"B_RAW_TYPE",						'RAWT'}, 
	{"B_RECT_TYPE",						'RECT'}, 
	{"B_REF_TYPE",						'RREF'}, 
	{"B_RGB_32_BIT_TYPE",				'RGBB'}, 
	{"B_RGB_COLOR_TYPE",				'RGBC'}, 
	{"B_SIZE_T_TYPE",					'SIZT'}, 
	{"B_SSIZE_T_TYPE",					'SSZT'}, 
	{"B_STRING_TYPE",					'CSTR'}, 
	{"B_TIME_TYPE",						'TIME'}, 
	{"B_UINT64_TYPE",					'ULLG'}, 
	{"B_UINT32_TYPE",					'ULNG'}, 
	{"B_UINT16_TYPE",					'USHT'}, 
	{"B_UINT8_TYPE",					'UBYT'}, 
	{"B_MEDIA_PARAMETER_TYPE",			'BMCT'}, 
	{"B_MEDIA_PARAMETER_WEB_TYPE",		'BMCW'}, 
	{"B_MEDIA_PARAMETER_GROUP_TYPE",	'BMCG'}, 
	{"B_ATOM_TYPE",						'ATOM'}, 
	{"B_ATOMREF_TYPE",					'ATMR'}, 
	{"B_ASCII_TYPE",					'TEXT'}
};

bool GetTypeCode(const char *text, type_code *code)
{
	if (text == NULL || code == NULL) return 0;
	if (strlen(text) == 0) return 0;

	unsigned long hex_value;
	char *hex_check;
	int i = 0;

	// Known type constant
	while (i < TYPE_CODES_COUNT) {
		if (strcasecmp(text, kTypeCodes[i].name) == 0) {
			*code = kTypeCodes[i].code;
			return true;
		}
		i++;
	}

	// Hex value
	if (strlen(text) < 3) return false;
	if (text[0] == '0' && (text[1] == 'X' || text[1] == 'x')) {
		if (strlen(text) > 10) return false;
		hex_value = strtoul(text + 2, &hex_check, 16);
		if (*hex_check != 0) return false;
		*code = static_cast<type_code>(hex_value);
		return true;
	}

	// Literal constants
	if (strlen(text) == 4) {
		memcpy(code, text, 4);
		*code = B_BENDIAN_TO_HOST_INT32(static_cast<uint32>(*code));
		return true;
	}

	return false;
}

bool GetTypeName(type_code code, char *buffer, unsigned bufSize)
{
	if (bufSize < 1 || buffer == NULL) return false;

	int i = 0;

	// Check if it's a knwon type
	while (i < TYPE_CODES_COUNT) {
		if (code == kTypeCodes[i].code)
			break;
		i++;
	}

	// It is a known type
	if (i < TYPE_CODES_COUNT) {
		if (strlen(kTypeCodes[i].name) >= bufSize)
			return false;
		strcpy(buffer, kTypeCodes[i].name);
		return true;
	}

	// Not a known type, return hex value
	if (bufSize < 11)
		return false;
	sprintf(buffer, "0x%08X", static_cast<unsigned int>(code));
	return true;
}

// Print help/list type ========================================================

void PrintTypes(void)
{
	int len, max_len = 0, i;
	char str[64], buffer[5];
	uint32 code;

	// Search for the longest type name (to make a column list)
	for (i = 0; i < TYPE_CODES_COUNT; i++) {
		len = strlen(kTypeCodes[i].name);
		if (len > max_len)
			max_len = len;
	}

	// The printf string. If max_len is 30, the resulting string will be
	// %-30s '%s' 0x%08X
	sprintf(str, "%%-%ds \'%%s\' 0x%%08X\n", max_len);

	// Print the list
	buffer[4] = 0;
	for (i = 0; i < TYPE_CODES_COUNT; i++) {
		code = B_HOST_TO_BENDIAN_INT32(static_cast<uint32>(kTypeCodes[i].code));
		memcpy(buffer, &code, 4);
		printf(str, kTypeCodes[i].name, buffer, kTypeCodes[i].code);
	}
}

void PrintUsage(void)
{
	puts(
		"Resource-To-File 1.1 by Gabriele Biffi - www.biffuz.it\n\n"
		"Extract a resource from a resource file and writes it to a file or "
		"stdout.\n\n"
		"Usage:\n\n"
		"restofile input_file -i id [-t type] [output_file]\n"
		"restofile -l\n\n"
		"\t-i id\tThe ID of the resource, a 32 bit integer. Can be decimal, "
		"hex\n\t\t(begins with 0x), or octal (begins with 0)\n"
		"\t-t type\tThe type of the resource, default is B_RAW_TYPE. Can be "
		"a\n\t\tconstant from TypeConstants.h, a 32 bit hex value (begins "
		"with\n\t\t0x), or a four characters literal constant.\n"
		"\t-l\tPrint the list of known types.\n\n"
		"Examples:\n\n"
		"\trestofile MyPictures.rsrc -i 0xFF -t B_RAW_TYPE 'My dog.jpg'\n\n"
		"\tExtract the resource 255 (0xFF in hex) of type B_RAW_TYPE from "
		"the\n\tresource file \"MyPictures.rsrc\" and saves it to the file \"My dog.jpg\".\n\n"	
		"\trestofile MyResource.rsrc -i 077 -t CSTR | gunzip -c > "
		"MyBook.txt\n\n"
		"\tExtract the previously compressed resource 63 (077 in octal) of "
		"type\n\tB_STRING_TYPE from \"MyResources.rsrc\" and uncompress it to "
		"the file\n\t\"MyBook.txt\".\n"
	);
}

// Main ========================================================================

int main(int argn, char *argv[])
{
	bool have_id, have_input, have_output, have_type, bad_args;
	char *input_file_name, *output_file_name, *id_check;
	BResources *res_file;
	BFile *input_file;
	FILE *output_file;
	const void *data;
	type_code code;
	size_t length;
	int32 id;
	int i;

	// If no arguments are given, print the help and exit
	if (argn < 2) {
		PrintUsage();
		return 0;
	}

	// Read the arguments
	i = 1;
	have_id = false;
	have_output = false;
	have_type = false;
	input_file_name = NULL;
	output_file_name = NULL;
	bad_args = false;
	while (i < argn && !bad_args) {
		if (strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0) {
			PrintUsage();
			return 0;
		}
		if (strcmp(argv[i], "-l") == 0) {
			PrintTypes();
			return 0;
		}
		if (strcmp(argv[i], "-i") == 0) {
			if (have_id || i == argn - 1) {
				bad_args = true;
				break;
			}
			have_id = true;
			id = static_cast<int32>(strtol(argv[++i], &id_check, 0));
			if (*id_check != 0) {
				bad_args = true;
				break;
			}
		}
		else if (strcmp(argv[i], "-t") == 0) {
			if (have_type || i == argn - 1) {
				bad_args = true;
				break;
			}
			have_type = true;
			if (!GetTypeCode(argv[++i], &code)) {
				bad_args = true;
				break;
			}
		}
		else {
			if (have_input) {
				if (have_output) {
					bad_args = true;
					break;
				}
				have_output = true;
				output_file_name = argv[i];
			}
			else {
				have_input = true;
				input_file_name = argv[i];
			}
		}
		i++;
	}

	#if 0 // Debug
	char buffer[64];
	GetTypeName(code, buffer, sizeof(buffer));
	printf("input = %s\noutput = %s\nid = %d\ncode = %s\n",
		input_file_name, output_file_name, static_cast<int>(id), buffer);
	return 0;
	#endif

	if (bad_args) {
		fprintf(stderr, "restofile: error: bad arguments\n");
		return -1;
	}

	// Check the arguments
	if (!have_id || !have_input) {
		fprintf(stderr, "restofile: error: missing arguments\n");
		return -1;
	}

	// Open resource file
	input_file = new BFile(input_file_name, B_READ_ONLY);
	if (input_file->InitCheck() != B_OK) {
		fprintf(stderr, "restofile: error: could not open input file %s\n",
			input_file_name);
		return -2;
	}

	// Initialize resources
	res_file = new BResources();
	if (res_file->SetTo(input_file) != B_OK) {
		fprintf(stderr, "restofile: error: input file %s is not a resources "
			"file\n", input_file_name);
		delete input_file;
		return -3;
	}
	delete input_file;	

	// Check type and ID
	if (!(res_file->HasResource(code, id))) {
		char buffer[64];
		GetTypeName(code, buffer, sizeof(buffer));
		fprintf(stderr, "restofile: error: can't find a resource of type %s "
			"and ID %d\n", buffer, static_cast<int>(id));
		delete res_file;
		return -4;
	}

	// Load data
	data = res_file->LoadResource(code, id, &length);
	if (data == NULL) {
		fprintf(stderr, "restofile: error: load resource failed\n");
		delete res_file;
		return -5;
	}

	// Open input file
	if (have_output) {
		output_file = fopen(output_file_name, "wb");
		if (output_file == NULL) {
			fprintf(stderr, "restofile: error: could not open output file %s\n",
				output_file_name);
			return -6;
		}
	}
	else
		output_file = stdout;

	// Write data
	if (fwrite(data, 1, length, output_file) != length) {
		fprintf(stderr, "restofile: error: write failed\n");
		if (have_output)
			fclose(output_file);
		delete res_file;
		return -7;
	}

	// Close files
	delete res_file;
	if (have_output)
		fclose(output_file);

	return 0;
}
